import { h } from '../rl-web-components.core.js';

class Coordinate {
    constructor(x, y) {
        this.x = x;
        this.y = y;
    }
    clone() {
        return new Coordinate(this.x, this.y);
    }
    toString() {
        return '(' + this.x + ', ' + this.y + ')';
    }
    toPathString() {
        return this.x + ' ' + this.y;
    }
    equals(other) {
        return other instanceof Coordinate && Coordinate.equals(this, other);
    }
    ceil() {
        this.x = Math.ceil(this.x);
        this.y = Math.ceil(this.y);
        return this;
    }
    floor() {
        this.x = Math.floor(this.x);
        this.y = Math.floor(this.y);
        return this;
    }
    translate(tx, ty) {
        if (tx instanceof Coordinate) {
            this.x += tx.x;
            this.y += tx.y;
        }
        else {
            this.x += Number(tx);
            if (typeof ty === 'number') {
                this.y += ty;
            }
        }
        return this;
    }
    scale(sx, sy) {
        sy = typeof sy === 'number' ? sy : sx;
        this.x *= sx;
        this.y *= sy;
        return this;
    }
    rotateRadians(radians, center) {
        center = center || new Coordinate(0, 0);
        const x = this.x;
        const y = this.y;
        const cos = Math.cos(radians);
        const sin = Math.sin(radians);
        this.x = (x - center.x) * cos - (y - center.y) * sin + center.x;
        this.y = (x - center.x) * sin + (y - center.y) * cos + center.y;
        return this;
    }
    rotateDegrees(degrees, center) {
        return this.rotateRadians(degrees * Math.PI / 180, center);
    }
    limit(limits) {
        if (!limits) {
            throw new Error('Unable to limit Corrdinate, limits undefined!');
        }
        const left = !isNaN(limits.left) ? limits.left : undefined;
        const width = !isNaN(limits.width) ? limits.width : 0;
        const maxX = left !== undefined ? left + width : Infinity;
        const minX = left !== undefined ? left : -Infinity;
        this.x = Math.min(maxX, Math.max(minX, this.x));
        const top = !isNaN(limits.top) ? limits.top : undefined;
        const height = !isNaN(limits.height) ? limits.height : 0;
        const maxY = top !== undefined ? top + height : Infinity;
        const minY = top !== undefined ? top : -Infinity;
        this.y = Math.min(maxY, Math.max(minY, this.y));
        return this;
    }
    round(precision = 0) {
        const factor = Math.pow(10, precision);
        this.x = Math.round(this.x * factor) / factor;
        this.y = Math.round(this.y * factor) / factor;
        return this;
    }
    static equals(a, b) {
        if (a === b) {
            return true;
        }
        return a.x === b.x && a.y === b.y;
    }
    static distance(a, b) {
        const dx = a.x - b.x;
        const dy = a.y - b.y;
        return Math.sqrt(dx * dx + dy * dy);
    }
    static magnitude(a) {
        return Math.sqrt(a.x * a.x + a.y * a.y);
    }
    static squareDistance(a, b) {
        const dx = a.x - b.x;
        const dy = a.y - b.y;
        return dx * dx + dy * dy;
    }
    static difference(a, b) {
        if (!b) {
            return new Coordinate(0, 0);
        }
        return new Coordinate(a.x - b.x, a.y - b.y);
    }
    static sum(a, b) {
        return new Coordinate(a.x + b.x, a.y + b.y);
    }
    static midpoint(a, b) {
        return new Coordinate((a.x + b.x) / 2, (a.y + b.y) / 2);
    }
}

function movePoints(c, dx, dy) {
    return c.map(i => i.translate(dx, dy));
}
function getClientPosition(el) {
    if (el.nodeType === 1) {
        return getClientPositionForElement_(el);
    }
    else {
        const targetEvent = el.changedTouches ? el.changedTouches[0] : el;
        return new Coordinate(targetEvent.clientX, targetEvent.clientY);
    }
}
function getClientPositionForElement_(el) {
    const box = getBoundingClientRect_(el);
    return new Coordinate(box.left, box.top);
}
function getBoundingClientRect_(el) {
    let rect;
    try {
        rect = el.getBoundingClientRect();
    }
    catch (e) {
        return { 'left': 0, 'top': 0, 'right': 0, 'bottom': 0 };
    }
    return rect;
}
function arrayReplace(array, index, item) {
    return [...array.slice(0, index), item, ...array.slice(index + 1)];
}
function coordinateFromEvent(a, b) {
    const ap = getClientPosition(a);
    const bp = getClientPosition(b);
    return new Coordinate(ap.x - bp.x, ap.y - bp.y);
}
function decodeCoordinates(data, closed) {
    const pointData = data.split(' ');
    const points = [];
    let i = pointData.length - 2;
    let x;
    let y;
    let path = closed ? 'Z' : '';
    while (i >= 0) {
        x = Math.round(Number(pointData[i]));
        y = Math.round(Number(pointData[i + 1]));
        points.push(new Coordinate(x, y));
        path = (i === 0 ? ['M', x, y, path].join(' ') : ['L', x, y, path].join(' '));
        i -= 2;
    }
    return { path, points };
}
function encodeCoordinates(points) {
    if (points instanceof Coordinate) {
        return points.toPathString();
    }
    else {
        return points.map(i => {
            return i.toPathString();
        }).join(' ');
    }
}
function pathFromCoordinateArray(points, open = false) {
    return points.map((i, idx) => {
        if (idx === 0) {
            return 'M ' + i.toPathString();
        }
        else if (idx === points.length - 1 && !open) {
            return ' L ' + i.toPathString() + ' Z';
        }
        else {
            return ' L ' + i.toPathString();
        }
    }).join();
}

class MapElement {
    constructor(id) {
        this.active = false;
        this.enabled = false;
        this.id = -1;
        this.name = 'New Element';
        this._coordinates = [];
        this.icons = [];
        this.scale = 1;
        if (id !== undefined) {
            this.id = id;
        }
    }
    move(delta) {
        this._coordinates = movePoints(this._coordinates, delta);
    }
}

const CONTROL_SIZE = 100;
class MapPoint extends MapElement {
    constructor() {
        super(...arguments);
        this.origin = new Coordinate(0, 0);
    }
    set coordinates(newCoordinates) {
        this._coordinates = newCoordinates;
        this.origin = newCoordinates[0];
    }
    get coordinates() {
        return [this.origin];
    }
    render() {
        if (!this.enabled) {
            return undefined;
        }
        const rectClass = {
            'rl-map-element': true,
            'rl-map-element__rect': true,
            'rl-map-element__rect--activated': this.active,
        };
        let contents = [];
        let totalWidth = 0;
        if (this.icons && this.icons.length !== 0) {
            contents = this.icons.map(icon => {
                const x = totalWidth;
                const scale = CONTROL_SIZE / (icon.width !== undefined ? icon.width : 1);
                totalWidth += icon.width !== undefined ? (icon.width * scale) : CONTROL_SIZE;
                return ([
                    h("image", { class: "rl-map-element__icon", height: CONTROL_SIZE, x: x, xlinkHref: icon.src }),
                ]);
            });
            let aspect = CONTROL_SIZE / totalWidth;
            aspect = isNaN(aspect) ? 1 : aspect;
            contents.push((h("rect", { class: rectClass, width: CONTROL_SIZE / aspect, height: CONTROL_SIZE, rx: "12", ry: "12" })));
        }
        else {
            totalWidth = CONTROL_SIZE;
            contents.push((h("rect", { class: rectClass, rx: "12", ry: "12", width: CONTROL_SIZE, height: CONTROL_SIZE })));
        }
        const gTrans = 'translate(' + (this.origin.x - totalWidth / 2) +
            ' ' + (this.origin.y - CONTROL_SIZE / 2) + ')';
        return (h("g", { id: this.id, class: "rl-map-element__point", tabindex: "0", transform: gTrans }, contents));
    }
    removePoint() {
    }
    addPoint(c, index) {
        this.origin = c;
        index = index;
    }
    renderControls() {
        return undefined;
    }
    movePoint(delta, index) {
        this.origin.translate(delta);
        index = index;
    }
}

class MapRegion extends MapElement {
    constructor() {
        super(...arguments);
        this.open = false;
        this.path = '';
    }
    set coordinates(newCoordinates) {
        this._coordinates = newCoordinates;
        this.path = pathFromCoordinateArray(newCoordinates);
    }
    get coordinates() {
        return this._coordinates;
    }
    render() {
        if (!this.enabled) {
            return undefined;
        }
        const regionClass = {
            'rl-map-element': true,
            'rl-map-element__region': true,
            'rl-map-element__region--activated': this.active,
        };
        return (h("path", { id: this.id, class: regionClass, d: this.path, tabindex: "0", "aria-label": this.name, role: "button" }));
    }
    renderControls() {
        const coords = this._coordinates;
        if (coords.length === 0) {
            return (h("g", null));
        }
        const points = coords.filter((_, idx) => (coords.length === 1 || (this.open && idx === coords.length - 1 ? false : true))).map((i, index) => {
            const controls = [(h("circle", { cx: i.x, cy: i.y, r: 8 / this.scale, index: index * 2, class: "rl-map-element__control" }))];
            if (!this.open) {
                const mid = (index === coords.length - 1) ? Coordinate.midpoint(i, coords[0]) :
                    Coordinate.midpoint(i, coords[index + 1]);
                controls.push((h("circle", { cx: mid.x, cy: mid.y, r: 8 / this.scale, index: index * 2 + 1, class: "rl-map-element__midpoint" })));
            }
            return controls;
        });
        const path = (h("path", { class: "rl-map-element__outline", d: pathFromCoordinateArray(coords, this.open) }));
        return (h("g", null,
            path,
            points));
    }
    addPoint(c, index) {
        const coords = this._coordinates;
        this.coordinates = index !== undefined ?
            [...coords.slice(0, index), c, ...coords.slice(index)] :
            [...coords, c];
    }
    removePoint(index) {
        if (index !== undefined) {
            this.coordinates.splice(index, 1);
        }
        else {
            this.coordinates.splice(this._coordinates.length - 1, 1);
        }
    }
    move(delta) {
        super.move(delta);
        this.path = pathFromCoordinateArray(this.coordinates);
    }
    movePoint(delta, index) {
        const c = this.coordinates;
        this.coordinates = arrayReplace(c, index, c[index].translate(delta));
    }
}

const HYSTERESIS = 8;
const DEFAULT_MAX_SCALE = 3;
const DEFAULT_MIN_SCALE = 1;
const STATES = {
    NORMAL: 0,
    GESTURE_DOWN: 1,
    DRAGGING: 2,
    ZOOMING: 3,
    ADD_REGION_INIT: 10,
    ADD_REGION_FIRST: 11,
    ADD_REGION: 12,
    ADD_POINT: 20,
    DRAG_ELEMENT: 30,
};
function computeLimits(iSize, sSize, scale) {
    return new DOMRect(Math.min(0, sSize.width - iSize.width * scale), Math.min(0, sSize.height - iSize.height * scale), Math.abs(iSize.width * scale - sSize.width), Math.abs(iSize.height * scale - sSize.height));
}
function getTargetId(el) {
    if (el instanceof HTMLImageElement || el instanceof SVGRectElement) {
        el = el.parentElement;
    }
    if (el instanceof SVGPathElement || el instanceof SVGGElement) {
        return Number(el.id);
    }
    return undefined;
}
function parseElements(elements, scale = 1) {
    return Object.keys(elements).map(i => {
        const el = elements[Number(i)];
        const decoded = decodeCoordinates(el.points, true);
        const parsedIcons = [];
        if (el.icons && el.icons.length !== 0) {
            el.icons.forEach((src, idx) => {
                const icon = new Image();
                icon.src = src;
                icon.onload = () => {
                    parsedIcons[idx] = {
                        src: icon.src,
                        width: icon.width,
                        height: icon.height,
                    };
                };
            });
        }
        const newEl = decoded.points.length === 1 ? new MapPoint() : new MapRegion();
        newEl.icons = parsedIcons;
        newEl.id = el.id;
        newEl.name = el.name;
        newEl.enabled = el.enabled;
        newEl.coordinates = decoded.points;
        newEl.scale = scale;
        return newEl;
    });
}

export { STATES as a, Coordinate as b, DEFAULT_MAX_SCALE as c, DEFAULT_MIN_SCALE as d, parseElements as e, coordinateFromEvent as f, getTargetId as g, HYSTERESIS as h, computeLimits as i, MapRegion as j, MapPoint as k, encodeCoordinates as l };
