const h = window.RlWebComponents.h;

/**
 * Class for representing coordinates and positions.
 * Adapted from Google Closure Library.
 */
class Coordinate {
    /**
     * Create a new Coordinate.
     */
    constructor(x, y) {
        this.x = x;
        this.y = y;
    }
    /**
     * Returns a new copy of the coordinate.
     */
    clone() {
        return new Coordinate(this.x, this.y);
    }
    /**
     * Returns a nice string representation the coordinate.
     */
    toString() {
        return '(' + this.x + ', ' + this.y + ')';
    }
    /**
     * Returns a basic string representation of the Coordinate used to easily
     * create SVG path strings.
     */
    toPathString() {
        return this.x + ' ' + this.y;
    }
    /**
     * Returns whether the specified value is equal to this coordinate.
     * @param other Some other value.
     */
    equals(other) {
        return other instanceof Coordinate && Coordinate.equals(this, other);
    }
    /**
     * Rounds the x and y fields to the next larger integer values.
     */
    ceil() {
        this.x = Math.ceil(this.x);
        this.y = Math.ceil(this.y);
        return this;
    }
    /**
     * Rounds the x and y fields to the next smaller integer values.
     */
    floor() {
        this.x = Math.floor(this.x);
        this.y = Math.floor(this.y);
        return this;
    }
    /**
     * Translates this coordinate by the given offsets. If a `Coordinate`
     * is given, then the x and y values are translated by the coordinate's x
     * and y. Otherwise, x and y are translated by `tx` and `ty` respectively.
     *
     * @param tx The value to translate x by or the
     *     the coordinate to translate this coordinate by.
     * @param ty The value to translate y by.
     */
    translate(tx, ty) {
        if (tx instanceof Coordinate) {
            this.x += tx.x;
            this.y += tx.y;
        }
        else {
            this.x += Number(tx);
            if (typeof ty === 'number') {
                this.y += ty;
            }
        }
        return this;
    }
    /**
     * Scales this coordinate by the given scale factors. The x and y values are
     * scaled by `sx` and `sy` respectively.  If `sy` is not given, then `sx` is
     * used for both x and y.
     *
     * @param sx The scale factor to use for the x dimension.
     * @param opt_sy The scale factor to use for the y dimension.
     */
    scale(sx, sy) {
        sy = typeof sy === 'number' ? sy : sx;
        this.x *= sx;
        this.y *= sy;
        return this;
    }
    /**
     * Rotates this coordinate clockwise about the origin (or, optionally, the
     * given center) by the given angle, in radians.
     *
     * @param radians The angle by which to rotate this coordinate
     *     clockwise about the given center, in radians.
     * @param center The center of rotation. Defaults
     *     to (0, 0) if not given.
     */
    rotateRadians(radians, center) {
        center = center || new Coordinate(0, 0);
        const x = this.x;
        const y = this.y;
        const cos = Math.cos(radians);
        const sin = Math.sin(radians);
        this.x = (x - center.x) * cos - (y - center.y) * sin + center.x;
        this.y = (x - center.x) * sin + (y - center.y) * cos + center.y;
        return this;
    }
    /**
     * Rotates this coordinate clockwise about the origin (or, optionally, the
     * given center) by the given angle, in degrees.
     *
     * @param degrees The angle by which to rotate this coordinate
     *     clockwise about the given center, in degrees.
     * @param opt_center The center of rotation. Defaults
     *     to (0, 0) if not given.
     */
    rotateDegrees(degrees, center) {
        return this.rotateRadians(degrees * Math.PI / 180, center);
    }
    /**
     * Restricts each parameter of the Coordinate such that
     * `left <= x <= left + width` and `top <= y <= top + height`.
     *
     * @param limits A rectangle to restrict the Coordinate to.
     */
    limit(limits) {
        // Abort if limits is undefined.
        if (!limits) {
            throw new Error('Unable to limit Corrdinate, limits undefined!');
        }
        // Limit in the x direction.
        const left = !isNaN(limits.left) ? limits.left : undefined;
        const width = !isNaN(limits.width) ? limits.width : 0;
        const maxX = left !== undefined ? left + width : Infinity;
        const minX = left !== undefined ? left : -Infinity;
        this.x = Math.min(maxX, Math.max(minX, this.x));
        // Limit in the y direction.
        const top = !isNaN(limits.top) ? limits.top : undefined;
        const height = !isNaN(limits.height) ? limits.height : 0;
        const maxY = top !== undefined ? top + height : Infinity;
        const minY = top !== undefined ? top : -Infinity;
        this.y = Math.min(maxY, Math.max(minY, this.y));
        return this;
    }
    /**
     * Round the values of the `Coordinate` to a given precision (number of
     * decimals).  Defaults to a precision of 0, or rounding to the nearest
     * integer.
     *
     * @param precision The number of decimals to round to
     */
    round(precision = 0) {
        const factor = Math.pow(10, precision);
        this.x = Math.round(this.x * factor) / factor;
        this.y = Math.round(this.y * factor) / factor;
        return this;
    }
    /**
     * Compares two `Coordinate`s for equlaity.
     * @param a The first `Coordinate`.
     * @param b The second `Coordinate`.
     */
    static equals(a, b) {
        if (a === b) {
            return true;
        }
        return a.x === b.x && a.y === b.y;
    }
    /**
     * Returns the distance between two `Coordinate`s.
     * @param a The first `Coordinate`.
     * @param b The second `Coordinate`.
     */
    static distance(a, b) {
        const dx = a.x - b.x;
        const dy = a.y - b.y;
        return Math.sqrt(dx * dx + dy * dy);
    }
    /**
     * Returns the magnitude (distance from origin) of a `Coordinate`.
     * @param a The first `Coordinate`.
     */
    static magnitude(a) {
        return Math.sqrt(a.x * a.x + a.y * a.y);
    }
    /**
     * Returns the squared distance between two `Coordinate`s. Squared distances
     * can be used for comparisons when the actual value is not required.
     *
     * Performance note: eliminating the square root is an optimization often used
     * in lower-level languages, but the speed difference is not nearly as
     * pronounced in JavaScript (only a few percent.)
     *
     * @param a The first `Coordinate`.
     * @param b The second `Coordinate`.
     */
    static squareDistance(a, b) {
        const dx = a.x - b.x;
        const dy = a.y - b.y;
        return dx * dx + dy * dy;
    }
    /**
     * Returns the difference between two `Coordinate`s.
     * @param a The first `Coordinate`.
     * @param b The second `Coordinate`.
     */
    static difference(a, b) {
        if (!b) {
            return new Coordinate(0, 0);
        }
        return new Coordinate(a.x - b.x, a.y - b.y);
    }
    /**
     * Returns the sum of two coordinates as a new `Coordinate`.
     * @param a The first `Coordinate`.
     * @param b The second `Coordinate`.
     */
    static sum(a, b) {
        return new Coordinate(a.x + b.x, a.y + b.y);
    }
    /**
     * Returns the midpoint of two coordinates as a new `Coordinate`.
     * @param a The first `Coordinate`.
     * @param b The second `Coordinate`.
     */
    static midpoint(a, b) {
        return new Coordinate((a.x + b.x) / 2, (a.y + b.y) / 2);
    }
}

/**
 * Moves a given array of `Coordinate`s by a specified distance in the
 * x and y axes.
 * @param c A `Coordinate` or `Coordinate`s to move.
 * @param dx A `Coordinate` to use for the translate values, or a number to move
 * in the x-axis;
 * @param dy The distance to move in the y-axis.
 */
function movePoints(c, dx, dy) {
    return c.map(i => i.translate(dx, dy));
}
/**
 * Returns the position of the event or the element's border box relative to
 * the client viewport. If an event is passed, and if this event is a "touch"
 * event, then the position of the first changedTouches will be returned.
 *
 * @param el Element or a mouse / touch event.
 * @return The position.
 */
function getClientPosition(el) {
    if (el.nodeType === 1 /* NodeType ELEMENT */) {
        return getClientPositionForElement_(el);
    }
    else {
        const targetEvent = el.changedTouches ? el.changedTouches[0] : el;
        return new Coordinate(targetEvent.clientX, targetEvent.clientY);
    }
}
/**
 * Returns the position of the event or the element's border box relative to
 * the client viewport.
 *
 * @param el Element whose position to get.
 * @return The position.
 */
function getClientPositionForElement_(el) {
    const box = getBoundingClientRect_(el);
    return new Coordinate(box.left, box.top);
}
/**
 * Gets the client rectangle of the DOM element.
 *
 * getBoundingClientRect is part of a new CSS object model draft (with a
 * long-time presence in IE), replacing the error-prone parent offset
 * computation and the now-deprecated Gecko getBoxObjectFor.
 *
 * This utility patches common browser bugs in getBoundingClientRect. It
 * will fail if getBoundingClientRect is unsupported.
 *
 * If the element is not in the DOM, the result is undefined, and an error may
 * be thrown depending on user agent.
 *
 * @param el The element whose bounding rectangle is being queried.
 * @return A native bounding rectangle with numerical left, top,
 *     right, and bottom.  Reported by Firefox to be of object type ClientRect.
 */
function getBoundingClientRect_(el) {
    let rect;
    try {
        rect = el.getBoundingClientRect();
    }
    catch (e) {
        // In IE < 9, calling getBoundingClientRect on an orphan element raises an
        // "Unspecified Error". All other browsers return zeros.
        return { 'left': 0, 'top': 0, 'right': 0, 'bottom': 0 };
    }
    return rect;
}
/**
 * Replaces an item in a given array with a new one.
 *
 * @param array The array to replace the item in.
 * @param index The index of the item to replace.
 * @param item The new item to place at `index`.
 */
function arrayReplace(array, index, item) {
    return [...array.slice(0, index), item, ...array.slice(index + 1)];
}
/**
 * Creates a new Coordinate based on the clientX/Y values of a given event or the
 * top-left corner of an element.  Optionally gets the position relative to a
 * second element.
 *
 * @param e The event to retrieve the client coordinates from.
 */
function coordinateFromEvent(a, b) {
    const ap = getClientPosition(a);
    const bp = getClientPosition(b);
    return new Coordinate(ap.x - bp.x, ap.y - bp.y);
}
/**
 * Parses a string of x/y coordinate pairs separated by spaces.  Returns an
 * object with two properties.  The `points` property is an array of Coordinate
 * objects representing each x/y pair and the `path` property is an SVG encoded
 * path string in the form `M x1 y1 L x2 y2 ... L xN yN Z`.
 *
 * @param data A string of space-separated coordinates.
 */
function decodeCoordinates(data, closed) {
    const pointData = data.split(' ');
    const points = [];
    let i = pointData.length - 2;
    let x;
    let y;
    // Closed paths end with 'Z'. Start with that and work backwards unless open.
    let path = closed ? 'Z' : '';
    while (i >= 0) {
        x = Math.round(Number(pointData[i]));
        y = Math.round(Number(pointData[i + 1]));
        points.push(new Coordinate(x, y));
        path = (i === 0 ? ['M', x, y, path].join(' ') : ['L', x, y, path].join(' '));
        i -= 2;
    }
    return { path, points };
}
/**
 * Turns an array of Coordinate objects into a single space-separated string.
 * @param points An array of Coordinates.
 */
function encodeCoordinates(points) {
    if (points instanceof Coordinate) {
        return points.toPathString();
    }
    else {
        return points.map(i => {
            return i.toPathString();
        }).join(' ');
    }
}
/**
 * Generates an SVG-style path string from an array of `Coordinates`.
 *
 * @param points An array of `Coordinates`.
 * @param open If the path represented by the array of points is open or closed.
 */
function pathFromCoordinateArray(points, open = false) {
    return points.map((i, idx) => {
        if (idx === 0) {
            return 'M ' + i.toPathString();
        }
        else if (idx === points.length - 1 && !open) {
            return ' L ' + i.toPathString() + ' Z';
        }
        else {
            return ' L ' + i.toPathString();
        }
    }).join();
}

class MapElement {
    /**
     * Create a new `MapElement`.
     * @param id A value to set as the ID of the `MapElement`.
     */
    constructor(id) {
        /**
         * A flag indicating if this `MapElement` is currently active.  The active
         * element is styled differently to make it stand out.
         */
        this.active = false;
        /**
         * A flag indicating if this `MapElement` is currently enabled.  An element
         * that is not enabled is not rendered on the Map.
         */
        this.enabled = false;
        /**
         * The identifier for this `MapElement`.
         */
        this.id = -1;
        /**
         * The name of this `MapElement`.
         */
        this.name = 'New Element';
        /**
         * The array of `Coordinate`s that are used to draw this `MapElement`.
         */
        this._coordinates = [];
        this.icons = [];
        this.scale = 1;
        if (id !== undefined) {
            this.id = id;
        }
    }
    /**
     * Move this `MapElement` by a specified amount.
     * @param delta The amount to move
     */
    move(delta) {
        this._coordinates = movePoints(this._coordinates, delta);
    }
}

const CONTROL_SIZE = 100;
class MapPoint extends MapElement {
    constructor() {
        super(...arguments);
        this.origin = new Coordinate(0, 0);
    }
    set coordinates(newCoordinates) {
        this._coordinates = newCoordinates;
        this.origin = newCoordinates[0];
    }
    get coordinates() {
        return [this.origin];
    }
    render() {
        if (!this.enabled) {
            return undefined;
        }
        const rectClass = {
            'rl-map-element': true,
            'rl-map-element__rect': true,
            'rl-map-element__rect--activated': this.active,
        };
        let contents = [];
        let totalWidth = 0;
        if (this.icons && this.icons.length !== 0) {
            contents = this.icons.map(icon => {
                const x = totalWidth;
                const scale = CONTROL_SIZE / (icon.width !== undefined ? icon.width : 1);
                totalWidth += icon.width !== undefined ? (icon.width * scale) : CONTROL_SIZE;
                return ([
                    h("image", { class: "rl-map-element__icon", 
                        // Don't set width.  Some icons are not square and leaving width
                        // unset will set the width automatically while keeping aspect.
                        height: CONTROL_SIZE, x: x, xlinkHref: icon.src }),
                ]);
            });
            let aspect = CONTROL_SIZE / totalWidth;
            aspect = isNaN(aspect) ? 1 : aspect;
            contents.push((h("rect", { class: rectClass, width: CONTROL_SIZE / aspect, height: CONTROL_SIZE, rx: "12", ry: "12" })));
        }
        else {
            totalWidth = CONTROL_SIZE;
            contents.push((h("rect", { class: rectClass, rx: "12", ry: "12", width: CONTROL_SIZE, height: CONTROL_SIZE })));
        }
        const gTrans = 'translate(' + (this.origin.x - totalWidth / 2) +
            ' ' + (this.origin.y - CONTROL_SIZE / 2) + ')';
        return (h("g", { id: this.id, class: "rl-map-element__point", tabindex: "0", transform: gTrans }, contents));
    }
    /**
     * An empty function that exists just to satisfy Typescript typings.
     */
    removePoint() {
        // No-op
    }
    /**
     * Adds a new point to this `MapElement`.  Since `MapPoint`s represent a
     * single point on a map, calling this function will simply set the current
     * position of the Point.
     * @param c The `Coordinate` that will get added
     * @param index Unused.
     */
    addPoint(c, index) {
        this.origin = c;
        index = index;
    }
    /**
     * Renders any elements needed to control the editing (moving) of this point.
     */
    renderControls() {
        return undefined;
    }
    /**
     * Moves a point by a given amount.  Since a `MapPoint` only has one point,
     * this function simply moves the origin by the specified amount.
     * @param delta The amount to move the point by
     * @param index Unused
     */
    movePoint(delta, index) {
        this.origin.translate(delta);
        index = index;
    }
}

class MapRegion extends MapElement {
    constructor() {
        super(...arguments);
        this.open = false;
        this.path = '';
    }
    set coordinates(newCoordinates) {
        this._coordinates = newCoordinates;
        this.path = pathFromCoordinateArray(newCoordinates);
    }
    get coordinates() {
        return this._coordinates;
    }
    /**
     * Renders the `MapRegion`.
     */
    render() {
        if (!this.enabled) {
            return undefined;
        }
        const regionClass = {
            'rl-map-element': true,
            'rl-map-element__region': true,
            'rl-map-element__region--activated': this.active,
        };
        return (h("path", { id: this.id, class: regionClass, d: this.path, tabindex: "0", "aria-label": this.name, role: "button" }));
    }
    /**
     * Generates the DOM necessary to display the control points for this
     * `MapElement`.  There are control points at the median and endpoint of each
     * edge.  This allows any point of the `MapElement` to be moved as well as new
     * point created on each line.
     */
    renderControls() {
        const coords = this._coordinates;
        if (coords.length === 0) {
            return (h("g", null));
        }
        const points = coords.filter((_, idx) => (coords.length === 1 || (this.open && idx === coords.length - 1 ? false : true))).map((i, index) => {
            const controls = [(h("circle", { cx: i.x, cy: i.y, r: 8 / this.scale, index: index * 2, class: "rl-map-element__control" }))];
            if (!this.open) {
                // Midpoint
                const mid = (index === coords.length - 1) ? Coordinate.midpoint(i, coords[0]) :
                    Coordinate.midpoint(i, coords[index + 1]);
                controls.push((h("circle", { cx: mid.x, cy: mid.y, r: 8 / this.scale, index: index * 2 + 1, class: "rl-map-element__midpoint" })));
            }
            return controls;
        });
        const path = (h("path", { class: "rl-map-element__outline", d: pathFromCoordinateArray(coords, this.open) }));
        return (h("g", null,
            path,
            points));
    }
    /**
     * Add a new point to the array of points that represent this `MapRegion`.
     * @param c The `Coordinate` representing the point to add
     * @param index The index where the point should be added.  If not specified,
     * adds the point at the end
     */
    addPoint(c, index) {
        const coords = this._coordinates;
        this.coordinates = index !== undefined ?
            [...coords.slice(0, index), c, ...coords.slice(index)] :
            [...coords, c];
    }
    /**
     * Remove a point from the array of points that represent this `MapRegion`.
     * If no index is specified, the last point is removed.
     * @param index The index of the point to remove
     */
    removePoint(index) {
        if (index !== undefined) {
            this.coordinates.splice(index, 1);
        }
        else {
            this.coordinates.splice(this._coordinates.length - 1, 1);
        }
    }
    /**
     * Move this `MapRegion` by a specified offset.
     * @param delta The amount to move by
     */
    move(delta) {
        super.move(delta);
        this.path = pathFromCoordinateArray(this.coordinates);
    }
    /**
     * Move a single point of this `MapRegion` by a specified amount.
     * @param delta The amount to move by
     * @param index The index of the point to move
     */
    movePoint(delta, index) {
        const c = this.coordinates;
        this.coordinates = arrayReplace(c, index, c[index].translate(delta));
    }
}

const CONTROL_SIZE$1 = 100;
const HYSTERESIS = 8;
// Default maximum scale (zoom) factor.
const DEFAULT_MAX_SCALE = 3;
// Default minimum scale (zoom) factor.
const DEFAULT_MIN_SCALE = 1;
// The different states the Map/MapEditor can be in.
const STATES = {
    NORMAL: 0,
    GESTURE_DOWN: 1,
    DRAGGING: 2,
    ZOOMING: 3,
    ADD_REGION_INIT: 10,
    ADD_REGION_FIRST: 11,
    ADD_REGION: 12,
    ADD_POINT: 20,
    DRAG_ELEMENT: 30,
};
const SVG_NS = 'http://www.w3.org/2000/svg';
function computeLimits(iSize, sSize, scale) {
    return new DOMRect(Math.min(0, sSize.width - iSize.width * scale), // left / x
    Math.min(0, sSize.height - iSize.height * scale), // top / y
    Math.abs(iSize.width * scale - sSize.width), // width
    Math.abs(iSize.height * scale - sSize.height) // height
    );
}
/**
 * Gets the ID of a given Element within the Map SVG.  This makes use of the
 * known DOM structure of the SVG.
 * @param el The Element to use to find the ID.
 */
function getTargetId(el) {
    if (el instanceof HTMLImageElement || el instanceof SVGRectElement) {
        // There are two cases where the element is a child of the elemen
        el = el.parentElement;
    }
    if (el instanceof SVGPathElement || el instanceof SVGGElement) {
        return Number(el.id);
    }
    return undefined;
}
/**
 * Processes an array of `MapElement`s, turning them into an array of
 * `ParsedMapElement` objects.  These parsed map elements are then rendered onto
 * a map.
 * @param elements An array of `MapElement`s to process.
 */
function parseElements(elements, scale = 1) {
    return Object.keys(elements).map(i => {
        const el = elements[Number(i)];
        const decoded = decodeCoordinates(el.points, true);
        const parsedIcons = [];
        if (el.icons && el.icons.length !== 0) {
            // Load all the icon images so they're cached later.
            el.icons.forEach((src, idx) => {
                const icon = new Image();
                icon.src = src;
                icon.onload = () => {
                    parsedIcons[idx] = {
                        src: icon.src,
                        width: icon.width,
                        height: icon.height,
                    };
                };
            });
        }
        const newEl = decoded.points.length === 1 ? new MapPoint() : new MapRegion();
        newEl.icons = parsedIcons;
        newEl.id = el.id;
        newEl.name = el.name;
        newEl.enabled = el.enabled;
        newEl.coordinates = decoded.points;
        newEl.scale = scale;
        return newEl;
    });
}
/**
 * Creates an array of JSX.Element nodes to be rendered based on an array of
 * `ParsedMapElements`.
 * @param els The array of `ParsedMapElements` to render
 */
function renderElements(els) {
    const elements = els.filter((el => el.enabled));
    if (elements.length === 0) {
        return undefined;
    }
    const parsed = elements.map(el => {
        if (el.coordinates.length === 1) {
            const rectClass = {
                'rl-map-element__rect': true,
                'rl-map-element__rect--activated': el.active,
            };
            const gTrans = 'translate(' + (el.coordinates[0].x - CONTROL_SIZE$1 / 2) +
                ' ' + (el.coordinates[0].y - CONTROL_SIZE$1 / 2) + ')';
            if (el.iconImages && el.iconImages.length !== 0) {
                return el.iconImages.map(icon => {
                    let iconAspect = icon.height / icon.width;
                    if (isNaN(iconAspect)) {
                        iconAspect = 1;
                    }
                    return (h("g", { id: el.id, class: "rl-map-element__point", tabindex: "0", transform: gTrans },
                        h("image", { class: "rl-map-element__icon", 
                            // Don't set width.  Some icons are not square and leaving width
                            // unset will set the width automatically while keeping aspect.
                            height: CONTROL_SIZE$1, xlinkHref: icon.src }),
                        h("rect", { class: rectClass, width: CONTROL_SIZE$1 / iconAspect, height: CONTROL_SIZE$1, rx: "12", ry: "12" })));
                });
            }
            else {
                // We have a text node.
                return (h("g", { id: el.id, class: "rl-map-element__point", tabindex: "0", transform: gTrans },
                    h("rect", { rx: "12", ry: "12", width: CONTROL_SIZE$1, height: CONTROL_SIZE$1 }))
                // <text id={el.id} class="rl-map-element__text" transform={gTrans}>
                // </text>
                );
            }
        }
        else {
            const regionClass = {
                'rl-map-element__region': true,
                'rl-map-element__region--activated': el.active,
            };
            return (h("path", { class: regionClass, id: el.id, d: el.path, tabindex: "0" }));
        }
    });
    return parsed;
}

export { STATES as a, Coordinate as b, DEFAULT_MAX_SCALE as c, DEFAULT_MIN_SCALE as d, parseElements as e, coordinateFromEvent as f, getTargetId as g, HYSTERESIS as h, computeLimits as i, MapRegion as j, MapPoint as k, encodeCoordinates as l };
